import json
from .controller import IController, TSettings
from ...deployment import IImageSource, IImageBuilder, Deployment, LocalImageSource, IContainerRunner, Command
import shutil
from .runner import BrainBoxRunner
from .run_configuration import RunConfiguration
import re
import os

class DockerController(IController[TSettings]):
    # region Abstract methods
    def get_image_source(self) -> IImageSource:
        return LocalImageSource(self.get_snake_case_name())

    def get_image_builder(self) -> IImageBuilder|None:
        return None

    # endregion

    def get_executor(self):
        return self.context.executor

    def get_deployment(self, runner: IContainerRunner|None = None):
        return Deployment(
            self.get_image_source(),
            runner,
            self.get_executor(),
            self.get_image_builder()
        )


    def find_self_in_list(self, all_images_installed: tuple[str,...]):
        image = self.get_image_source().get_image_name().split(':')[0]
        if image in all_images_installed:
            return image
        else:
            return None


    def install(self):
        self.context.logger.log("Running pre-install")
        self.pre_install()
        self.context.logger.log("Stopping the current container, if exists")
        self.get_deployment().stop()
        self.context.logger.log("Removing the current container, if persists")
        self.get_deployment().remove()
        self.context.logger.log("Building/pulling image")
        self.get_deployment().obtain_image()
        self.context.logger.log("Running post-install")
        self.post_install()
        self.context.logger.log("DONE")

    def pre_install(self):
        pass

    def post_install(self):
        pass

    def uninstall(self, purge: bool = False):
        self.get_deployment().stop().remove()
        images = self.get_image_source().get_relevant_images(self.get_executor())
        for image in images:
            self.get_executor().execute(['docker', 'rmi', image])
        if purge:
            for filename in os.listdir(self.resource_folder()):
                file_path = self.resource_folder()/filename
                if file_path.is_file():
                    os.unlink(file_path)  # Remove file or symlink
                if file_path.is_dir():
                    shutil.rmtree(file_path)  # Remove directory and its contents
            shutil.rmtree(self.resource_folder())

    def run_with_configuration(self, configuration: RunConfiguration, monitor_function = print) -> str:
        command = configuration.generate_command(
            self.get_image_source().get_container_name(),
            self.get_image_source().get_image_name(),
            self.context
        )
        print(" ".join(command))
        result = self.get_executor().execute(command, Command.Options(monitor_output=monitor_function))
        return result[:12]

    def stop(self, instance_id: str):
        self.get_executor().execute(['docker','stop',instance_id], Command.Options(ignore_exit_code=True))
        self.get_executor().execute(['docker', 'rm', instance_id], Command.Options(ignore_exit_code=True))




    def instance_id_to_parameter(self, instance_id):
        parameter = self.get_executor().execute(
            ['docker', 'exec', instance_id, 'printenv', 'BRAINBOX_PARAMETER'],
            Command.Options(return_output=True)
        )
        parameter = parameter.strip()
        if parameter.startswith('*'):
            parameter = parameter[1:]
        else:
            parameter = None
        return parameter

    def get_running_instances_id_to_parameter(self) -> dict[str, str|None]:
        containers = self.get_executor().execute(
            [
                'docker', 'ps',
                '--format', '{"id":"{{.ID}}","name":"{{.Names}}"}',
                '--filter', f'ancestor={self.get_image_source().get_image_name()}'
            ],
            Command.Options(return_output=True)
        )
        result = {}
        containers = containers.strip()
        if containers == '':
            return result
        for container in containers.split('\n'):
            data = json.loads(container)
            result[data['id']] = self.instance_id_to_parameter(data['name'])
        return result

    def run_auxiliary_configuration(self, cfg):
        runner = BrainBoxRunner(self.context, cfg, False)
        runner.run(
            self.get_image_source().get_image_name(),
            self.get_image_source().get_container_name(),
            self.get_executor()
        )





from ..core import Core, ICoreAction, JobForPlanner, Job
from sqlalchemy import select
from datetime import datetime



class TaskForPlannerSyncWithTracking(ICoreAction):
    def __init__(self):
        self.last_time_full_update: datetime|None = None


    def should_do_full_update(self, core):
        if self.last_time_full_update is None:
            return True
        if core.jobs_for_planner is None:
            return True
        if (datetime.now() - self.last_time_full_update).total_seconds() > 5:
            return True
        if len(core.jobs_for_planner) < 100: #This is the optimization for really big loads. Doesn't make sense to do it for smaller
            return True
        return False

    def apply(self, core: Core):
        with core.new_session() as session:
            if self.should_do_full_update(core):
                self.last_time_full_update = datetime.now()
                core.jobs_for_planner = self._get_tasks(session)
                return

            changed = set(core.new_session.changes.all)
            not_changed = tuple(
                job
                for job in core.jobs_for_planner
                if job.id not in changed
            )
            core.jobs_for_planner = not_changed + self._get_tasks(session, set(j.id for j in not_changed))
            core.new_session.reset()



    def _get_tasks(self, session, skip_for_ids: None|set = None):
        condition = ~Job.finished & Job.ready
        if skip_for_ids is not None:
            condition &= Job.id.notin_(skip_for_ids)
        new_records = list(session.execute(
            select(
                Job.id,
                Job.decider,
                Job.decider_parameter,
                Job.received_timestamp,
                Job.assigned,
                Job.ordering_token
            )
            .where(condition)
        ))
        return tuple(JobForPlanner(*rec) for rec in new_records)



from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import PiperModel

@dataclass
class PiperSettings:
    connection = ConnectionSettings(20205)
    models_to_download = (
        PiperModel(
            name="en",
            config_url="https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/en/en_GB/alan/medium/en_GB-alan-medium.onnx.json?download=true",
            url="https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/en/en_GB/alba/medium/en_GB-alba-medium.onnx"
        ),
        PiperModel(
            name="ru",
            config_url='https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/ru/ru_RU/ruslan/medium/ru_RU-ruslan-medium.onnx.json?download=true.json',
            url='https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/ru/ru_RU/ruslan/medium/ru_RU-ruslan-medium.onnx?download=true'
        ),
        PiperModel(
            name="de",
            config_url="https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/de/de_DE/thorsten/medium/de_DE-thorsten-medium.onnx.json?download=true.json",
            url="https://huggingface.co/rhasspy/piper-voices/resolve/v1.0.0/de/de_DE/thorsten/medium/de_DE-thorsten-medium.onnx?download=true"
        )
    )


from dataclasses import dataclass
from ....framework import ConnectionSettings


@dataclass
class ModelInstallationSettings:
    model_name: str
    mode: str = 'simple'
    test_dub: bool = True
    test_voice_clone: bool = False

    @property
    def due_folder_name(self):
        return self.model_name.replace('/','--')


@dataclass
class CoquiTTSSettings:
    connection = ConnectionSettings(20202)
    builtin_models_to_download = (
        ModelInstallationSettings('tts_models/en/vctk/vits'),
        ModelInstallationSettings('tts_models/multilingual/multi-dataset/your_tts'),
        ModelInstallationSettings('tts_models/multilingual/multi-dataset/xtts_v2', mode='say_yes',test_dub=False, test_voice_clone=True),
    )







from brainbox.deciders import WhisperX
from brainbox.framework import Loc
import dotenv

if __name__ == '__main__':
    dotenv.load_dotenv(Loc.root_folder/'environment.env')
    controller = WhisperX.Controller()
    controller.install()
    controller.self_test()


from ....framework import DockerWebServiceApi, FileLike, ISingleLoadableModelApi
import requests
from .settings import VoskSettings
from .controller import VoskController
from .model import VoskModel


class Vosk(DockerWebServiceApi[VoskSettings, VoskController], ISingleLoadableModelApi):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def load_model(self, model_name):
        result = requests.post(self.endpoint('load_model'), json=dict(model=model_name))
        if result.status_code != 200:
            raise ValueError(f"Endpoint load_model returned exception\n{result.text}")


    def get_loaded_model_name(self) -> str|None:
        result = requests.get(self.endpoint('get_loaded_model'))
        if result.status_code != 200:
            raise ValueError(f"Endpoint returned exception\n{result.text}")
        return result.json()


    def transcribe(self, file: FileLike.Type, model_name: str|None = None):
        with FileLike(file, self.cache_folder) as stream:
            ep = 'transcribe'
            if model_name is not None:
                ep+='/'+model_name
            result = requests.post(
                self.endpoint('transcribe/'+model_name),
                files=(
                    ('file', stream),
                ))
            if result.status_code != 200:
                raise ValueError(f"Endpoint transcribe returned error\n{result.text}")
            return result.json()

    def transcribe_to_array(self, file: FileLike.Type, model_name: str|None = None):
        result = self.transcribe(file, model_name)
        return [item for part in result for item in part['result']]

    def transcribe_to_string(self, file: FileLike.Type, model_name: str|None = None):
        return ' '.join(z['word'] for z in self.transcribe_to_array(file, model_name))


    Settings = VoskSettings
    Controller = VoskController
    Model = VoskModel



from brainbox.deciders.text.espeak_phonemizer.controller import EspeakPhonemizerController

if __name__ == '__main__':
    controller = EspeakPhonemizerController()
    controller.install()
    #controller.run_notebook()
    controller.self_test()


import traceback

from .command import  Command
import subprocess
from dataclasses import dataclass
import threading


@dataclass
class ExecutionScenario:
    command: tuple[str,...]
    options: Command.Options
    command_as_str: str
    call_kwargs: dict


    _monitored_output = None

    def _to_str(self, output):
        if isinstance(output, str):
            return output
        try:
            return output.decode('utf-8')
        except:
            return output

    def _execute_with_no_output_control(self):
        try:
            result = subprocess.call(self.command, **self.call_kwargs)
        except Exception as ex:
            raise ValueError(f"Error launching subprocess, arguments:\n{self.command_as_str}") from ex
        if result != 0:
            if self.options.ignore_exit_code:
                pass
            else:
                raise ValueError(f'Error running subprocess, arguments:\n{self.command_as_str}')
        return None

    def _execute_with_return_output(self):
        try:
            output = subprocess.check_output(self.command, **self.call_kwargs, stderr=subprocess.STDOUT)
            return self._to_str(output)
        except subprocess.CalledProcessError as exc:
            o_str = self._to_str(exc.output)
            if self.options.ignore_exit_code:
                return o_str
            else:
                raise ValueError(f'Error running subprocess, arguments:\n{self.command_as_str}\noutput\n{o_str}')


    def _monitor(self, stream):
        """Reads lines from the stream and calls the callback with each line."""
        try:
            for line in iter(stream.readline, ""):
                if line=='':
                    break
                no_newline = line
                if no_newline.endswith('\n'):
                    no_newline = no_newline[:-1]
                self.options.monitor_output(no_newline)
                self._monitored_output.append(no_newline)

            stream.close()
        except:
            ex = traceback.format_exc()
            self.options.monitor_output("Error when reading stream. The process continues to run, but the monitoring is broken")
            self.options.monitor_output(ex)


    def _execute_with_monitoring(self):
        self._monitored_output = []
        try:
            process = subprocess.Popen(
                self.command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                **self.call_kwargs,
                text=True,
                bufsize=1,
                encoding='utf-8'
            )

            stdout_thread = threading.Thread(target=self._monitor, args=(process.stdout,), daemon=True)
            stderr_thread = threading.Thread(target=self._monitor, args=(process.stderr,), daemon=True)

            stdout_thread.start()
            stderr_thread.start()

            process.wait()

            stdout_thread.join()
            stderr_thread.join()

        except Exception as ex:
            raise ValueError(f"Error launching subprocess, arguments:\n{self.command_as_str}") from ex

        if process.returncode != 0:
            if not self.options.ignore_exit_code:
                raise ValueError(f'Error running subprocess, arguments:\n{self.command_as_str}\noutput\n'+"\n".join(self._monitored_output[-50:]))

        return "\n".join(self._monitored_output)



    def  execute(self):
        if self.options.monitor_output is not None:
            return self._execute_with_monitoring()
        elif self.options.return_output:
            return self._execute_with_return_output()
        else:
            return self._execute_with_no_output_control()






from ..core import ICoreAction, Core, Job
from sqlalchemy.orm import Session
from sqlalchemy import select
from datetime import datetime
from dataclasses import dataclass

@dataclass
class CancelAction(ICoreAction):
    cancel_job_id: str|None
    cancel_batch_id: str|None

    def apply(self, core: Core):
        condition = ~Job.finished
        if self.cancel_job_id is not None:
            condition &= Job.id == self.cancel_job_id
        if self.cancel_batch_id is not None:
            condition &= Job.batch == self.cancel_batch_id
        with core.new_session() as session:
            jobs = list(session.scalars(select(Job).where(condition)))
            for job in jobs:
                core.close_job(session, job, 'Cancelled')
            session.commit()


from ..brainbox_mapping import BrainBoxMapping
from foundation_kaia.prompters import IPrompter
from ..applicator import IApplicator
from ...parsers import IParser
from ..object_to_task import PromptBasedObjectConverter
from brainbox import BrainBox
from .step import IStep, IStepFactory
from pathlib import Path

class BrainBoxMappingStep(IStep):
    def __init__(self, api: BrainBox.Api, mapping: BrainBoxMapping, cache: Path|None = None):
        self.api = api
        self.mapping = mapping
        self.cache = cache

    def process(self, history, current):
        return self.mapping.create(current, self.cache).execute(self.api)

    def shorten(self, data):
        return data.result


class PromptMappingStepFactory(IStepFactory):
    def __init__(self, api: BrainBox.Api, prompter: IPrompter, model: str, applicator: IApplicator, parser: IParser|None = None):
        self.api = api
        self.prompter = prompter
        self.model = model
        self.applicator = applicator
        self.parser = parser

    def create_step(self) -> IStep:
        return BrainBoxMappingStep(
            self.api,
            BrainBoxMapping(
                PromptBasedObjectConverter(
                    self.prompter,
                    self.model,
                ),
                self.applicator,
                self.parser
            )
        )


from brainbox.deciders import Resemblyzer
from unittest import TestCase

if __name__ == '__main__':
    controller = Resemblyzer.Controller()
    controller.install()
    controller.self_test()
    #controller.run_notebook()




from typing import *
from dataclasses import dataclass, field
from enum import Enum
from .decider_instance_key import DeciderInstanceKey
from datetime import datetime

@dataclass
class OperatorLogItem:
    class Level(Enum):
        Service = 0
        Decider = 1
        Task = 2
        Core = 3
        Planer = 4

    level: 'OperatorLogItem.Level'
    event: str
    id: Optional[str] = None
    error: Optional[str] = None
    timestamp: datetime = field(default_factory=datetime.now)

    def __str__(self):
        id_width = 35  # Fixed width for id
        level_width = 10  # Fixed width for type

        id = ''
        if self.id is not None:
            id = self.id

        level = ''
        if self.level is not None:
            level = self.level.name
        details = f"{level:<{level_width}} {id:<{id_width}}  {self.event}"

        return details

@dataclass
class OperatorLogHandle:
    callback: Optional[Callable]
    level: OperatorLogItem.Level
    _id: Optional[str]

    def event(self, s: str):
        if self.callback is not None:
            self.callback(OperatorLogItem(self.level, s, self._id))

    def error(self, traceback: str, error_message='Error'):
        if self.callback is not None:
            self.callback(OperatorLogItem(self.level, error_message, self._id, traceback))

@dataclass
class OperatorLog:
    debug_output: bool = False
    log_items: list[OperatorLogItem] = field(default_factory = list)

    def _append(self, item):
        if self.debug_output:
            print(item)
        self.log_items.append(item)
        if len(self.log_items) > 5000:
            self.log_items = self.log_items[2500:]

    def service(self):
        return OperatorLogHandle(self._append, OperatorLogItem.Level.Service, None)

    def decider(self, name: DeciderInstanceKey):
        return OperatorLogHandle(self._append, OperatorLogItem.Level.Decider, str(name))

    def task(self, id: str):
        return OperatorLogHandle(self._append, OperatorLogItem.Level.Task, id)

    def core(self):
        return OperatorLogHandle(self._append, OperatorLogItem.Level.Core, None)

    def planer(self):
        return OperatorLogHandle(self._append, OperatorLogItem.Level.Planer, None)





from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import RhasspyKaldiModel

@dataclass
class RhasspyKaldiSettings:
    connection = ConnectionSettings(20101)
    languages: tuple[str,...] = (
        RhasspyKaldiModel('en'),
        RhasspyKaldiModel('de'),
    )



from typing import Optional, Callable
from ...common import IDecider
from typing import Generic, TypeVar
from .run_configuration import RunConfiguration


TSettings = TypeVar('TSettings')
TController = TypeVar('TController')


class MonitorFunctionWrapper:
    def __init__(self, monitor_function):
        self.monitor_function = monitor_function

    def __call__(self, line):
        print(line)
        if self.monitor_function is not None:
            self.monitor_function(line)

class OnDemandDockerApi(IDecider, Generic[TSettings, TController]):
    @property
    def controller(self) -> TController:
        return self._controller

    @controller.setter
    def controller(self, controller: TController):
        self._controller = controller

    def run_container(self, configuration: RunConfiguration, monitor_function: Optional[Callable[[str],None]] = None):
        from .on_demand_docker_controller import OnDemandDockerController
        controller: OnDemandDockerController = self.controller
        controller.get_deployment().stop().remove()
        controller.run_with_configuration(configuration, MonitorFunctionWrapper(monitor_function))






from brainbox.deciders.images.video_to_images.controller import VideoToImagesController

if __name__ == '__main__':
    controller = VideoToImagesController()
    controller.install()
    controller.self_test()







import requests
from ....framework import DockerWebServiceApi, FileLike
from .controller import RhasspyKaldiController
from .settings import RhasspyKaldiSettings
from .model import RhasspyKaldiModel

class RhasspyKaldi(DockerWebServiceApi[RhasspyKaldiSettings, RhasspyKaldiController]):
    def __init__(self, address: str | None = None):
        super().__init__(address)

    def train(self, model: str, language: str, sentences: str, custom_words: dict[str, str]|None = None):
        if custom_words is not None:
            custom_words_str = '\n'.join(f'{k} {v}' for k, v in custom_words.items())
        else:
            custom_words_str = ''
        reply = requests.post(f'http://{self.address}/train/{language}/{model}', json=dict(sentences=sentences, custom_words=custom_words_str))
        if reply.status_code != 200:
            raise ValueError(f"RhasspyKaldi couldn't train for {language}/{model}\n{reply.text}")
        return reply.json()

    def transcribe(self, file: FileLike.Type, model: str):
        with FileLike(file, self.cache_folder) as file:
            reply = requests.post(
                f'http://{self.address}/transcribe/{model}',
                files=(
                    ('file', file),
                )
            )
            if reply.status_code != 200:
                raise ValueError(f"RhasspyKaldi couldn't transcribe for {model}\n{reply.text}")
            return reply.json()

    def phonemes(self, language: str):
        return requests.get(f'http://{self.address}/phonemes/{language}').text

    Controller = RhasspyKaldiController
    Settings = RhasspyKaldiSettings
    Model = RhasspyKaldiModel

    @classmethod
    def get_ordering_arguments_sequence(cls) -> tuple[str,...]|None:
        return ('model',)


'''

`[order]
_order = Shi|Ukha|Borsh
I order <_order>{_order}`
`shi S i
ukha u h 'A
borsh b o r S`

'''


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel
)
from .settings import EspeakPhonemizerSettings
from pathlib import Path


class EspeakPhonemizerController(
    DockerWebServiceController[EspeakPhonemizerSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return EspeakPhonemizerSettings()

    def create_api(self):
        from .api import EspeakPhonemizer
        return EspeakPhonemizer()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import EspeakPhonemizer

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize("This is some text"))

        yield (
            TestReport
            .last_call(api)
            .href('en-pho')
            .with_comment("Phonemizes one english line")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize(["This is some text", "This is the second line"]))

        yield (
            TestReport
            .last_call(api)
            .href('en-pho-2')
            .with_comment("Phonemizes several english lines")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize("Keep the stress marks", stress=True))
        yield (
            TestReport
            .last_call(api)
            .href('stress')
            .with_comment("Phonemizes keeping the stress marks")
        )

        api.execute(BrainBoxTask.call(EspeakPhonemizer).phonemize('Das is ein Satz', language='de'))
        yield (
            TestReport
                .last_call(api)
                .href('de')
                .with_comment("Phonemizes German")
        )





DOCKERFILE = f'''
FROM python:3.8.20

{SmallImageBuilder.APT_INSTALL('libespeak-ng1')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

RUN pip freeze

COPY . /home/app/

ENTRYPOINT ["python3","/home/app/main.py"]
'''

DEPENDENCIES = '''
anyio==4.5.2
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==3.0.0
async-lru==2.0.4
attrs==24.3.0
babel==2.16.0
backcall==0.2.0
beautifulsoup4==4.12.3
bleach==6.1.0
blinker==1.8.2
certifi==2024.12.14
cffi==1.17.1
charset-normalizer==3.4.1
click==8.1.8
comm==0.2.2
debugpy==1.8.11
decorator==5.1.1
defusedxml==0.7.1
espeak-phonemizer==1.3.1
exceptiongroup==1.2.2
executing==2.1.0
fastjsonschema==2.21.1
Flask==3.0.3
fqdn==1.5.1
h11==0.14.0
httpcore==1.0.7
httpx==0.28.1
idna==3.10
importlib_metadata==8.5.0
importlib_resources==6.4.5
ipykernel==6.29.5
ipython==8.12.3
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.2
Jinja2==3.1.5
json5==0.10.0
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2023.12.1
jupyter-events==0.10.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.14.2
jupyter_server_terminals==0.5.3
jupyterlab==4.3.4
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
MarkupSafe==2.1.5
matplotlib-inline==0.1.7
mistune==3.1.0
nbclient==0.10.1
nbconvert==7.16.5
nbformat==5.10.4
nest-asyncio==1.6.0
notebook==7.3.2
notebook_shim==0.2.4
overrides==7.7.0
packaging==24.2
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
pickleshare==0.7.5
pkgutil_resolve_name==1.3.10
platformdirs==4.3.6
prometheus_client==0.21.1
prompt_toolkit==3.0.48
psutil==6.1.1
ptyprocess==0.7.0
pure_eval==0.2.3
pycparser==2.22
Pygments==2.19.1
python-dateutil==2.9.0.post0
python-json-logger==3.2.1
pytz==2024.2
PyYAML==6.0.2
pyzmq==26.2.0
referencing==0.35.1
requests==2.32.3
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rpds-py==0.20.1
Send2Trash==1.8.3
six==1.17.0
sniffio==1.3.1
soupsieve==2.6
stack-data==0.6.3
terminado==0.18.1
tinycss2==1.2.1
tomli==2.2.1
tornado==6.4.2
traitlets==5.14.3
types-python-dateutil==2.9.0.20241206
typing_extensions==4.12.2
uri-template==1.3.0
urllib3==2.2.3
wcwidth==0.2.13
webcolors==24.8.0
webencodings==0.5.1
websocket-client==1.8.0
Werkzeug==3.0.6
zipp==3.20.2
'''




from .step import IStep
from typing import Any, Callable, Optional
from dataclasses import dataclass
from abc import ABC, abstractmethod
import numpy as np

@dataclass
class AbstractRepresentationStep(IStep, ABC):
    count: int
    item_to_category: Callable[[Any], Any]
    item_filter: Optional[Callable[[Any], bool]] = None
    item_to_count: Optional[Callable[[Any], int]] = None

    def _add_item_to_dict(self, item, d):
        category = self.item_to_category(item)
        if self.item_to_count is None:
            delta = 1
        else:
            delta = self.item_to_count(item)
        d[category] = d.get(category, 0) + delta

    def process(self, history: list, current: list) -> tuple[list, dict]:
        if len(current) <= self.count:
            selected = current
        else:
            categories_count = {}
            for item in history:
                if self.item_filter is not None and not self.item_filter(item):
                    continue
                self._add_item_to_dict(item, categories_count)
            selected = self.select(current, categories_count)

        categories_count = {}
        for item in selected:
            self._add_item_to_dict(item, categories_count)
        return (selected, categories_count)

    def shorten(self, data):
        return data[0]

    def summarize(self, data) -> str|None:
        return ', '.join(f'{k}: {v}' for k, v in data[1].items())


    @abstractmethod
    def select(self, current, categories_count) -> list:
        pass


def get_random_weighed_element(elements) -> int:
    csum = []
    for element in elements:
        if len(csum) == 0:
            csum.append(element)
        else:
            csum.append(csum[-1]+element)
    rnd = np.random.rand()*csum[-1]
    for i, cs in enumerate(csum):
        if cs > rnd:
            return i
    return len(csum) - 1


@dataclass
class ProbabilisticEqualRepresentationStep(AbstractRepresentationStep):
    magnitude: float = 1

    def select(self, current, categories_count) -> list:
        if len(current)<self.count:
            return current

        weights = []
        for item in current:
            category = self.item_to_category(item)
            w = 1 / ( 1+ categories_count.get(category, 0)) ** self.magnitude
            weights.append(w)

        seen = set()
        output = []
        for i in range(self.count):
            while True:
                idx = get_random_weighed_element(weights)
                if idx in seen:
                    continue
                seen.add(idx)
                weights[idx] = 0
                output.append(current[idx])
                break

        return output











from typing import Iterable
from unittest import TestCase

from . import TestReport
from ...common import IDecider, ISelfManagingDecider
from .controller import IController, TSettings
from uuid import uuid4


class ControllerOverDecider(IController):
    def __init__(self,
                 decider: IDecider,
                 custom_name: str|None = None
                 ):
        self.decider = decider
        self.custom_name = custom_name
        self.virtual_instances = {}

    def install(self):
        pass

    def uninstall(self, purge: bool = False):
        pass

    def find_self_in_list(self, all_images_installed: tuple[str,...]) -> str|None|IController.DockerlessInstallation:
        return IController.DockerlessInstallation()

    def get_default_settings(self) -> TSettings:
        return object()

    def run(self, parameter: str|None) -> str:
        if isinstance(self.decider, ISelfManagingDecider):
            self.decider.warmup(parameter)
        instance_id = str(uuid4())
        self.virtual_instances[instance_id] = parameter
        return instance_id

    def stop(self, instance_id: str):
        if isinstance(self.decider, ISelfManagingDecider):
            return self.decider.cooldown()
        del self.virtual_instances[instance_id]

    def get_running_instances_id_to_parameter(self) -> dict[str, str|None]:
        return self.virtual_instances

    def get_name(self):
        if self.custom_name is not None:
            return self.custom_name
        return type(self.decider).__name__

    def find_api(self, instance_id: str):
        return self.decider

    def _self_test_internal(self, tc: TestCase) -> Iterable:
        pass






import base64
from pathlib import Path
import requests
from ....framework import DockerWebServiceApi, FileLike
from .controller import YoloController
from .settings import YoloSettings

class Yolo(DockerWebServiceApi):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def analyze(self, file: FileLike.Type, model: str = None):
        loaded_model = self.get_loaded_model()
        if model is None:
            if loaded_model is None:
                raise ValueError("No model is provided and no model is preloaded")
        else:
            if loaded_model is None or loaded_model != model:
                self.load_model(model)

        with FileLike(file, self.cache_folder) as stream:
            data = base64.b64encode(stream.read()).decode("utf-8")
            response = requests.post(f"http://{self.address}/post_image", data={"image_base64": data})
        if response.status_code != 200:
            raise ValueError(response.text)
        response = requests.get(f"http://{self.address}/get_coordinates_faces")
        if response.status_code != 200:
            raise ValueError(response.text)
        return response.json()


    def load_model(self, model_id: str):
        reply = requests.post(
            f"http://{self.address}/load_model",
            json={
            "model_id": model_id,
        })
        if reply.status_code != 200:
            raise ValueError(reply.text)


    def get_loaded_model(self):
        reply = requests.get(f"http://{self.address}/get_loaded_model")
        return reply.json()['name']


    Controller = YoloController
    Settings = YoloSettings


from typing import Optional,Callable,Any
from abc import ABC, abstractmethod
from ..app import BrainBoxApi
from dataclasses import dataclass
from pathlib import Path

class IPostprocessor(ABC):
    @abstractmethod
    def postprocess(self, result, api):
        pass

class DefaultPostprocessor(IPostprocessor):
    def postprocess(self, result, api):
        return result

@dataclass
class DownloadingPostprocessorBase(IPostprocessor):
    take_element_before_downloading: Optional[int] = None

    def postprocess(self, result, api):
        if self.take_element_before_downloading is not None:
            try:
                result = result[self.take_element_before_downloading]
            except Exception as ex:
                raise ValueError(f"Cannot take element {self.take_element_before_downloading} from value {result}") from ex
        return result


_MISSING = object()

@dataclass
class FilePostprocessor(DownloadingPostprocessorBase):
    metadata: Any = _MISSING

    def postprocess(self, result, api):
        result = super().postprocess(result, api)
        file = api.open_file(result)
        if self.metadata is not _MISSING:
            file.metadata = self.metadata
        return file


@dataclass
class DownloadingPostprocessor(DownloadingPostprocessorBase):
    opener: Optional[Callable[[Path], Any]] = None

    def postprocess(self, result, api: BrainBoxApi):
        result = super().postprocess(result, api)
        result = api.download(result)
        if self.opener is not None:
            result = self.opener(result)
        return result



from .api import ControllerApi
from .interface import IControllerService
from .server import ControllerServer, ControllerServerSettings
from .service import ControllerService
from .dto import ControllersSetup, ControllerInstanceSetup, ControllerServiceStatus



import os
import shutil

from brainbox.framework import Loc
from brainbox.framework.deployment import SmallImageBuilder, LocalExecutor
from pathlib import Path


def create_brainbox_builder():
    bb_folder = Path(__file__).parent.parent.parent
    build_folder = Loc.temp_folder/'brain_box_container'
    if build_folder.is_dir():
        shutil.rmtree(build_folder)
    os.makedirs(build_folder)
    return SmallImageBuilder(
        build_folder,
        TEMPLATE,
        [d for d in DEPENDENCIES.split('\n') if d.strip()!=''],
        add_current_user=True,
        copy_to_code_path={
            bb_folder.parent/'foundation_kaia': '/foundation_kaia',
            bb_folder/'framework' : '/brainbox/framework',
            bb_folder/'deciders': '/brainbox/deciders',
            bb_folder/'run.py': '/brainbox/run.py',
            bb_folder/'__init__.py': '/brainbox/__init__.py',
            },
        write_to_code_path={
            '/pyproject.toml': TOML
        }
    )

DOCKER_INSTALL = '''
RUN apt-get update \
    && apt-get install -y --no-install-recommends \
       ca-certificates \
       curl \
       gnupg \
       lsb-release \
    && curl -fsSL https://download.docker.com/linux/debian/gpg | gpg --dearmor -o /usr/share/keyrings/docker-archive-keyring.gpg \
    && echo "deb [arch=$(dpkg --print-architecture) signed-by=/usr/share/keyrings/docker-archive-keyring.gpg] https://download.docker.com/linux/debian $(lsb_release -cs) stable" > /etc/apt/sources.list.d/docker.list \
    && apt-get update \
    && apt-get install -y docker-ce-cli \
    && rm -rf /var/lib/apt/lists/*
'''


TEMPLATE = f'''
FROM python:3.11.11

{DOCKER_INSTALL}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

COPY --chown=app:app . /home/app

RUN pip install --user -e .

ENTRYPOINT ["python", "-m", "brainbox.run"]
'''

TOML = '''
[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"

[project]
version = "0.0.0"
name = "brainbox-container"

[tool.setuptools]
packages = ["brainbox", "foundation_kaia"]


'''

DEPENDENCIES = """
asttokens==3.0.0
blinker==1.9.0
certifi==2024.12.14
charset-normalizer==3.4.1
click==8.1.8
comm==0.2.2
contourpy==1.3.1
cycler==0.12.1
decorator==5.1.1
executing==2.1.0
Flask==3.1.0
fonttools==4.55.3
greenlet==3.1.1
idna==3.10
ipython==8.31.0
ipywidgets==8.1.5
itsdangerous==2.2.0
jedi==0.19.2
Jinja2==3.1.5
joblib==1.4.2
jsonpickle==4.0.1
jupyterlab_widgets==3.0.13
kiwisolver==1.4.8
MarkupSafe==3.0.2
matplotlib==3.10.0
matplotlib-inline==0.1.7
numpy==2.2.1
packaging==24.2
pandas==2.2.3
parso==0.8.4
pexpect==4.9.0
pillow==11.0.0
prompt_toolkit==3.0.48
ptyprocess==0.7.0
pure_eval==0.2.3
pyaml==24.12.1
Pygments==2.18.0
pyparsing==3.2.0
python-dateutil==2.9.0.post0
pytz==2024.2
PyYAML==6.0.2
requests==2.32.3
scikit-learn==1.6.0
scipy==1.14.1
seaborn==0.13.2
six==1.17.0
SQLAlchemy==2.0.36
stack-data==0.6.3
threadpoolctl==3.5.0
toml==0.10.2
tqdm==4.67.1
traitlets==5.14.3
typing_extensions==4.12.2
tzdata==2024.2
urllib3==2.3.0
wcwidth==0.2.13
Werkzeug==3.1.3
widgetsnbextension==4.0.13
yo_fluq_ds==2.0.5
"""

if __name__ == '__main__':
    create_brainbox_builder().build_image('brainbox', LocalExecutor())


from .rhasspy_kaldi import RhasspyKaldi
from .whisper import Whisper
from .resemblyzer import Resemblyzer
from .resemble_enhance import ResembleEnhance
from .vosk import Vosk
from .whisperx import WhisperX



from .docker_web_service_api import DockerWebServiceApi
from .connection_settings import ConnectionSettings
from .controller import IController
from .controller_context import ControllerContext
from .docker_controller import DockerController
from .docker_web_service_controller import DockerWebServiceController
from .resource_folder import ResourceFolder
from .run_configuration import RunConfiguration
from .runner import BrainBoxRunner
from .test_report import TestReport, create_self_test_report_page
from .controller_over_decider import ControllerOverDecider
from .controller_registry import ControllerRegistry, InstallationStatus
from .notebookable_controller import INotebookableController
from .single_loadable_model_api import ISingleLoadableModelApi
from .model_downloading_controller import IModelDownloadingController, DownloadableModel
from .on_demand_docker_api import OnDemandDockerApi
from .on_demand_docker_controller import OnDemandDockerController


import traceback

import flask
from pathlib import Path
import shutil
import os
import pickle
import uuid
import traceback
from model import Model, Transcriber
from utils import profile_path

DATASETS_PATH = Path('/data/datasets')
MODELS_FOLDER = Path('/data/models')

class RhasspyKaldiServer:
    def __init__(self):
        self.transcribers: dict[str, Transcriber] = {}

    def __call__(self):
        app = flask.Flask(__name__)
        app.add_url_rule('/train/<language>/<model>', view_func=self.train, methods=['POST'])
        app.add_url_rule('/transcribe/<model>', view_func=self.transcribe, methods=['POST'])
        app.add_url_rule('/phonemes/<language>', view_func=self.phonemes, methods=['GET'])
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.run('0.0.0.0',8084)

    def index(self):
        return 'OK'

    def transcribe(self, model):
        try:
            file = flask.request.files['file']
            content = file.stream.read()
            if model not in self.transcribers:
                self.transcribers[model] = Model(model).create_transcriber()
            return flask.jsonify(self.transcribers[model].transcribe(content))
        except:
            return traceback.format_exc(), 500


    def train(self, language, model):
        try:
            js = flask.request.json
            sentences = js['sentences']
            custom_words = js['custom_words']
            model_object = Model(model)
            result = model_object.train(language, sentences, custom_words)
            self.transcribers[model] = model_object.create_transcriber()
            return flask.jsonify(result)
        except:
            return traceback.format_exc(), 500

    def phonemes(self, language):
        with open(profile_path/language/'kaldi/phoneme_examples.txt') as file:
            return file.read()










from brainbox.deciders.voice_analysis.faster_whisper import FasterWhisper

if __name__ == '__main__':
    controller = FasterWhisper.Controller()
    controller.install()
    controller.run_notebook()
    #controller.self_test()



from .planner_action import IPlannerAction
from ..core import DeciderInstanceKey, Core, OperatorState, Job
from ..operator import DeciderOperator
from threading import Thread
from ...common import IDecider
from sqlalchemy import select
import traceback
from dataclasses import dataclass
from enum import Enum

@dataclass
class StartCommand(IPlannerAction):
    class Mode(Enum):
        StartOnly = 0
        FindThenStart = 1
        FindOnly = 2

    key: DeciderInstanceKey
    mode: 'StartCommand.Mode'


    def _start(self, controller):
        instance_id = controller.run(self.key.parameter)
        api: IDecider = controller.find_api(instance_id)
        return instance_id, api

    def _use_existing_or_start(self, controller):
        instances = controller.get_running_instances_id_to_parameter()
        instance_id = None
        api = None
        for key, value in instances.items():
            if value == self.key.parameter:
                instance_id = key
                api = controller.find_api(instance_id)

        if instance_id is not None and api is not None:
            return instance_id, api

        if self.mode == StartCommand.Mode.FindOnly:
            if instance_id is None:
                raise ValueError(f"Could not find {self.key} among running instances for {self.key}")
            if api is None:
                raise ValueError(f"Could not find for {self.key}, instance {instance_id}")

        return self._start(controller)


    def apply(self, core: Core):
        core.operator_log.decider(self.key).event(f"Setting up container in the mode {self.mode.name}")
        try:
            controller = core.registry.get_controller(self.key.decider_name)
            if self.mode == StartCommand.Mode.StartOnly:
                instance_id, api = self._start(controller)
            else:
                instance_id, api = self._use_existing_or_start(controller)

            api.context._cache_folder = core.locator.cache_folder
            op_state = OperatorState(
                self.key,
                controller,
                instance_id,
                api,
                core.operator_log,
            )
            operator = DeciderOperator(op_state)
            thread = Thread(target=operator.cycle)
            op_state.operator_thread = thread
            thread.start()
            core.operator_states[instance_id] = op_state
            core.operator_log.decider(self.key).event("Initialized")
        except:
            core.operator_log.decider(self.key).event("Initialization failed")
            with core.new_session() as session:
                tasks: list[Job] = list(session.scalars(
                    select(Job)
                    .where(
                        ~Job.finished &
                        (Job.decider == self.key.decider_name) &
                        (Job.decider_parameter == self.key.parameter)
                    )
                ))
                for task in tasks:
                    if task.get_key() == self.key:
                        core.close_job(session, task, f'Decider {self.key} failed to warm up:\n{traceback.format_exc()}')
                session.commit()


from typing import TypeVar, Generic

TKey = TypeVar('TKey')

class ICacheManager(Generic[TKey]):
    def contains(self, key: TKey) -> bool:
        return False

    def unsafe_get(self, key: TKey):
        raise ValueError(f"Object {key} was not found")

    def get(self, key: TKey):
        if not self.contains(key):
            return None
        return self.unsafe_get(key)

    def store(self, key: TKey, value):
        pass

    def unsafe_delete(self, key: TKey):
        raise ValueError(f"Object {key} was not found")

    def delete(self, key: TKey):
        if self.contains(key):
            self.unsafe_delete(key)








import pickle
import traceback
import flask
import os
from zonos_model import ZonosModel
from pathlib import Path
import subprocess

def concat_with_mpeg(temp_file, output_path, paths):
    control_content_array = [f"file '{p}'" for p in paths]
    control_content = '\n'.join(control_content_array)
    with open(temp_file, 'w') as stream:
        stream.write(control_content)

    args = [
        'ffmpeg',
        '-f',
        'concat',
        '-safe',
        '0',
        '-i',
        str(temp_file),
        '-y',
        str(output_path)
    ]
    try:
        subprocess.check_output(args)
    except subprocess.CalledProcessError as err:
        raise ValueError(f'FFMpeg returned non-zero value. arguments are\n{" ".join(args)}. Output\n{err.output}')


class ZonosApp:
    def __init__(self, model: ZonosModel):
        self.model = model
        self.loaded_speakers = dict()

    def create_app(self):
        app = flask.Flask(type(self).__name__)
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/train/<speaker>', view_func=self.train, methods = ['POST'])
        app.add_url_rule('/voiceover', view_func=self.voiceover, methods=['POST'])
        return app

    def index(self):
        return "OK"

    def train(self, speaker: str):
        tmp_file = Path('temp.wav')
        try:
            source_folder = Path(f'/voices/{speaker}/')
            paths = [source_folder / f for f in os.listdir(source_folder)]
            concat_with_mpeg('files.tmp', tmp_file, paths)

            speaker_model = self.model.train(tmp_file)
            with open(f'/speakers/{speaker}', 'wb') as file:
                pickle.dump(speaker_model, file)
            self.loaded_speakers[speaker] = speaker_model
            return "OK"
        except:
            return traceback.format_exc(), 500
        finally:
            if tmp_file.is_file():
                os.unlink(tmp_file)

    def voiceover(self):
        output_file = Path(__file__).parent/'output.wav'
        try:
            data: dict = flask.request.json
            speaker = data['speaker']
            if speaker not in self.loaded_speakers:
                speaker_path = Path(f'/speakers/{speaker}')
                if not speaker_path.is_file():
                    raise ValueError(f"Speaker {speaker} was not trained")
                with open(speaker_path, 'rb') as stream:
                    self.loaded_speakers[speaker] = pickle.load(stream)
            text = data['text']
            language = data['language']
            emotion = data['emotion']
            speaking_rate = data['speaking_rate']
            self.model.voiceover(text, self.loaded_speakers[speaker], language, output_file, emotion, speaking_rate)
            return flask.send_file(output_file)
        except:
            return traceback.format_exc(), 500
        finally:
            if output_file.is_file():
                os.unlink(output_file)



from typing import *
from .test_report_item import TestReportItem, TestReportItemHolder
from .test_report_section import TestReportSection
from .main_section_content import MainSectionContent
from .source_file_attachment import SourceFileAttachment
import inspect
from yo_fluq import FileIO


def make_hierarchy(name: str, items: list) -> list[TestReportSection]:
    sections = [TestReportSection(name)]
    for item in items:
        if isinstance(item, TestReportSection):
            sections.append(item)
        elif isinstance(item, MainSectionContent):
            sections[0].comment = item.comment
        elif isinstance(item, TestReportItemHolder):
            sections[-1].items.append(item.item)
        elif isinstance(item, SourceFileAttachment):
            pass
        else:
            raise ValueError(f"Unexpected type of item in report\n{item}")
    return sections


def assemble_code_files(items: list, controller_type: Type):
    source_codes = []
    for item in items:
        if isinstance(item, SourceFileAttachment):
            source_codes.append(FileIO.read_text(inspect.getfile(item.method_or_type)))
    if len(source_codes) == 0:
        source_codes.append(inspect.getsource(getattr(controller_type,'_self_test_internal')))
    return source_codes




class TestReport:
    def __init__(self,
                 name: str,
                 items: Iterable[Union[TestReportItemHolder, TestReportSection]],
                 error: str|None,
                 controller_type: Type
                 ):
        items = list(items)
        self.name = name
        self.sections: list[TestReportSection] = make_hierarchy(name, items)
        self.error = error
        self.source_codes = assemble_code_files(items, controller_type)


    @staticmethod
    def last_call(brain_box_api) -> TestReportItemHolder:
        from ....brainbox import BrainBoxApi
        api: BrainBoxApi = brain_box_api
        summary = api.summary()
        recent_id = list(sorted(summary, key=lambda z: z['received_timestamp'], reverse=True))[0]['id']
        job = api.job(recent_id)
        return TestReportItemHolder(
            api,
            TestReportItem(
                job.decider,
                job.method,
                job.decider_parameter,
                job.arguments,
                job.result,
                job.accepted_timestamp,
                job.finished_timestamp,
        ))

    @staticmethod
    def section(caption: str, comment: str|None = None):
        return TestReportSection(caption, comment)

    @staticmethod
    def main_section_content(comment: str):
        return MainSectionContent(comment)

    @staticmethod
    def attach_source_file(method_or_class: Any):
        return SourceFileAttachment(method_or_class)





from typing import TypeVar, Generic
from .docker_web_service_controller import DockerWebServiceController
from ...common import ApiUtils, IDecider

TSettings = TypeVar('TSettings')
TController = TypeVar('TController')

class DockerWebServiceApi(IDecider, Generic[TSettings, TController]):
    def __init__(self, address: str|None, container_parameter: str|None = None):
        if address is not None:
            ApiUtils.check_address(address)
        self._custom_address = address
        self._controller: DockerWebServiceController|None = None
        self._container_parameter: container_parameter



    @property
    def controller(self) -> TController:
        return self._controller

    @controller.setter
    def controller(self, controller: TController):
        self._controller = controller

    @property
    def address(self) -> str:
        if self._custom_address is not None:
            return self._custom_address
        return self.controller.address

    def endpoint(self, endpoint=''):
        if len(endpoint)>0 and not endpoint.startswith('/'):
            endpoint='/'+endpoint
        return 'http://'+self.address+endpoint

    @property
    def settings(self) -> TSettings:
        return self.controller.settings

    @property
    def container_parameter(self) -> str|None:
        return self._container_parameter



import os
import shutil
import uuid
from typing import *
from pathlib import Path
from .image_builder import IImageBuilder
from ..executor import IExecutor, Command
from yo_fluq import FileIO


user_add_template = '''
RUN useradd -ms /bin/bash app -u {user_id}

RUN usermod -aG sudo app

USER app
'''

class SmallImageBuilder(IImageBuilder):
    def __init__(self,
                 code_path: Path,
                 docker_template: str|None,
                 dependencies: Optional[Iterable[str]|Iterable[Iterable[str]]],
                 add_current_user: bool = True,
                 copy_to_code_path: dict[Path, str]|None = None,
                 write_to_code_path: dict[str, str]|None = None,
                 reset_code_folder: bool = False,
                 build_command: Optional[Iterable[str]] = None
    ):
        self.code_path = code_path
        self.docker_template = docker_template
        self.dependencies = self._make_dependencies(dependencies)
        self.add_current_user = add_current_user
        self.copy_to_code_path = copy_to_code_path
        self.reset_code_folder = reset_code_folder
        self.write_to_code_path = write_to_code_path
        self.build_command = build_command

    ADD_USER_PLACEHOLDER = 'add_user'

    PIP_INSTALL_PLACEHOLDER = 'pip_install'

    def _make_dependencies(self, dependencies) -> list[list[str]]|None:
        if dependencies is None:
            return None
        deps = list(dependencies)
        if all(isinstance(element,str) for element in deps):
            deps = [deps]
        else:
            deps = [list(element) for element in dependencies]
        return deps

    def _convert_dependencies(self):
        if self.dependencies is None:
            return None
        result = []
        for line in self.dependencies:
            template = (
                'RUN pip wheel --no-cache-dir --wheel-dir=/tmp/wheels {DEPS} && '  
                'pip install --no-cache-dir --no-index --find-links=/tmp/wheels {DEPS} && '
                'rm -rf /tmp/wheels'
            )
            result.append(template.format(DEPS=' '.join(line)))
        return '\n\n'.join(result)

    def _create_target_file(self, target_str_path: str):
        target = target_str_path
        while target.startswith('/'):
            target = target[1:]
        target = self.code_path / target
        target.relative_to(self.code_path)
        if target.is_file():
            os.unlink(target)
        if target.is_dir():
            shutil.rmtree(target, ignore_errors=True)
        return target

    def _prepare_container_folder(self, executor: IExecutor):
        if self.docker_template is not None:
            format_kwargs = {}
            if self.dependencies is not None:
                format_kwargs[SmallImageBuilder.PIP_INSTALL_PLACEHOLDER] = self._convert_dependencies()
            if self.add_current_user:
                format_kwargs[SmallImageBuilder.ADD_USER_PLACEHOLDER] = user_add_template.format(user_id=executor.get_machine().user_id, group_id=executor.get_machine().group_id)

            dockerfile = self.docker_template.format(**format_kwargs)
            if self.reset_code_folder:
                shutil.rmtree(self.code_path, ignore_errors=True)
            os.makedirs(self.code_path, exist_ok=True)
            with open(self.code_path / 'Dockerfile', 'w', encoding='utf-8') as file:
                file.write(dockerfile)

        if self.copy_to_code_path is not None:
            for source_path, target_str_path in self.copy_to_code_path.items():
                target = self._create_target_file(target_str_path)
                if source_path.is_file():
                    os.makedirs(target.parent, exist_ok=True)
                    shutil.copy(source_path, target)
                elif source_path.is_dir():
                    shutil.copytree(source_path, target)
                else:
                    raise ValueError(f"{source_path} is neither file nor directory")

        if self.write_to_code_path is not None:
            for target_str_path, content in self.write_to_code_path.items():
                target = self._create_target_file(target_str_path)
                FileIO.write_text(content, target)



    def build_image(self, image_name: str, executor: IExecutor) -> None:
        self._prepare_container_folder(executor)
        args = []
        command = ['docker', 'build']
        if self.build_command is not None:
            command = ['docker'] + list(self.build_command)
        executor.execute(
            command + ['-t', image_name] + args + ['.'],
            Command.Options(workdir=self.code_path)
        )

    @staticmethod
    def APT_INSTALL(libs: str):
        return (
            'RUN apt-get update && '
            f'apt-get install -y --no-install-recommends {libs} && '
            'rm -rf /var/lib/apt/lists/*'
        )


    @staticmethod
    def GIT_CLONE_AND_RESET(repo, folder, commit=None, options='', install=True):
        result = []
        result.append(f'RUN git clone {repo} {folder}')
        result.append(f' && cd {folder}')
        if commit is not None:
            result.append(f' && git reset --hard {commit}')
        result.append(' && rm -rf .git')
        if install:
            result.append(f' && pip install --user -e .{options}')
            result.append(' && rm -rf ~/.cache/pip')
        return ''.join(result)



from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import OllamaModel

@dataclass
class OllamaSettings:
    connection = ConnectionSettings(20401)
    models_to_install: tuple[OllamaModel,...] = (
        OllamaModel('llama3.2:1b'),
    )



import pickle
import sys
import flask
from pathlib import Path
import os
import traceback
from generate import Generator
from tortoise.api import TextToSpeech
import sys
import argparse
import subprocess
from io import BytesIO

DEFAULT_MODELS_DIR = os.path.join(os.path.expanduser('~'), '.cache', 'tortoise', 'models')
MODELS_DIR = os.environ.get('TORTOISE_MODELS_DIR', DEFAULT_MODELS_DIR)


generator = None
app = flask.Flask("TortoiseTTSApp")


@app.route('/', methods=['GET'])
def index():
    return 'OK'



@app.route('/dub', methods=['POST'])
def dub():
    try:
        data = flask.request.json
        voice = data['voice']
        text = data['text']
        count = int(data['count'])
        result = generator.simple(text, voice, count)
        js = flask.jsonify(result)
        return js
    except:
        return traceback.format_exc(), 500




if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--install', action='store_true')
    parser.add_argument('-n','--notebook', action='store_true')
    args = parser.parse_args()

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0', "--NotebookApp.token=''"], cwd='/repo')
        exit(0)

    tts = TextToSpeech()
    if args.install:
        exit(0)

    generator = Generator(tts)
    app.run(port = 8084, host='0.0.0.0')








from brainbox.framework.deployment import IContainerRunner, DockerArgumentsHelper, Deployment, LocalImageSource, LocalExecutor
from dataclasses import dataclass
from brainbox.framework import Loc, IExecutor, Command
from pathlib import Path
from .container_builder import create_brainbox_builder

@dataclass
class BrainBoxRunner(IContainerRunner):
    folder: Path
    port: int = 18090
    auto_restart: bool = False
    user_id: str|None = None
    docker_group_id: str|None = None
    debug: bool = False


    def _get_user(self, executor: IExecutor):
        if self.user_id is not None:
            user_id = self.user_id
        else:
            user_id = int(executor.execute(['id','-u'], Command.Options(return_output=True)).strip())
        if self.docker_group_id is not None:
            docker_group_id = self.docker_group_id
        else:
            docker = executor.execute(['getent', 'group', 'docker'], Command.Options(return_output=True)).strip()
            docker_group_id = int(docker.split(':')[2])
        return ['--user', f'{user_id}:{docker_group_id}']


    def run(self, image_name: str, container_name: str, executor: IExecutor):
        restart = ['--restart', 'unless-stopped'] if self.auto_restart and not self.debug else []


        command = [
            'docker',
            'run',
            '--name',
            container_name,
            '--mount',
            DockerArgumentsHelper.arg_mount(self.folder, self.folder),
            '--network',
            'host',
            *(['--detach'] if not self.debug else ['--rm']),
            '-v',
            '/var/run/docker.sock:/var/run/docker.sock',
            *self._get_user(executor),
            *restart,
            image_name,
            '--data-folder',
            str(self.folder),
            '--port',
            str(self.port)
        ]
        executor.execute(command)

    def get_deployment(self):
        return Deployment(
            LocalImageSource('brainbox'),
            self,
            LocalExecutor(),
            create_brainbox_builder()
        )






from ...framework import IDecider, File
import json
from copy import copy
import time

class FakeFile(IDecider):
    def __call__(self, tags: dict=None, extension_with_dot='', array_length: int|None = None, time_to_sleep: float|None = None):
        if time_to_sleep is not None:
            time.sleep(time_to_sleep)
        if tags is None:
            tags = {}
        if array_length is None:
            return File(self.current_job_id+'.output'+extension_with_dot, json.dumps(tags))
        else:
            result = []
            for i in range(array_length):
                new_tags = copy(tags)
                new_tags['option_index'] = i
                result.append(File(self.current_job_id+'.output.'+str(i)+extension_with_dot, json.dumps(new_tags)))
            return result





from foundation_kaia.marshalling import ApiUtils, SignatureProcessor
from .decider import IDecider, DeciderContext, ISelfManagingDecider
from .file import File
from .file_like import FileLike
from foundation_kaia.fork import Fork
from .loc import Loc, Locator
from .logger import Logger
from .file_io import FileIO
from .HTML import HTML


from typing import *
from abc import ABC, abstractmethod
from ..executor import IExecutor

from pathlib import Path
import os
from enum import Enum

class IContainerRunner(ABC):
    @abstractmethod
    def run(self, image_name: str, container_name: str, executor: IExecutor):
        pass




class DockerArgumentsHelper:
    @staticmethod
    def arg_mount(host_path, container_path):
        return f'type=bind,source={host_path},target={container_path}'

    @staticmethod
    def arg_publish_ports(publish_ports: None | dict[int, int]) -> list[str]:
        arguments = []
        for host_port, container_port in publish_ports.items():
            arguments.extend(['--publish', f'{host_port}:{container_port}'])
        return arguments

    @staticmethod
    def arg_mount_folders(mount_folders: None | dict[str | Path, str]) -> list[str]:
        arguments = []
        if mount_folders is not None:
            for host_path, container_path in mount_folders.items():
                arguments.extend(['--mount', DockerArgumentsHelper.arg_mount(host_path, container_path)])
        return arguments

    @staticmethod
    def arg_propagate_env_variables(env_variables: Optional[Iterable[str]]) -> list[str]:
        arguments = []
        if env_variables is not None:
            for variable in env_variables:
                arguments.append('--env')
                arguments.append(f'{variable}={os.environ[variable]}')
        return arguments

    @staticmethod
    def arg_env_variables(*propagate, **set):
        return DockerArgumentsHelper.arg_propagate_env_variables(propagate) + DockerArgumentsHelper.arg_set_env_variables(set)


    @staticmethod
    def arg_set_env_variables(env_variables: None | dict[str, str]):
        arguments = []
        if env_variables is not None:
            for variable, value in env_variables.items():
                arguments.append('--env')
                arguments.append(f'{variable}={value}')
        return arguments



import pickle
import traceback
from typing import *
from abc import ABC, abstractmethod
from unittest import TestCase
from ...common import Locator, Loc
from .resource_folder import ResourceFolder
from .controller_context import ControllerContext
from .test_report import create_self_test_report_page, TestReport
from yo_fluq import FileIO


TSettings = TypeVar("TSettings")


class IController(ABC, Generic[TSettings]):
    # region Abstract methods

    @abstractmethod
    def get_default_settings(self) -> TSettings:
        pass

    @abstractmethod
    def install(self):
        pass

    @abstractmethod
    def uninstall(self, purge: bool = False):
        pass

    class DockerlessInstallation:
        pass

    @abstractmethod
    def find_self_in_list(self, all_images_installed: tuple[str,...]) -> Union[str, None, 'IController.DockerlessInstallation']:
        pass



    @abstractmethod
    def run(self, parameter: str|None) -> str:
        pass

    @abstractmethod
    def stop(self, instance_id: str):
        pass


    @abstractmethod
    def get_running_instances_id_to_parameter(self) -> dict[str, str|None]:
        pass

    @abstractmethod
    def find_api(self, instance_id: str):
        pass

    @abstractmethod
    def _self_test_internal(self, api, tc: TestCase) -> Iterable:
        pass


    #endregion
    # region Context properties

    @property
    def context(self) -> ControllerContext[TSettings]:
        if not hasattr(self, '_context'):
            self._context = ControllerContext(self.get_name(), self.get_default_settings())
        return self._context

    @property
    def resource_folder(self) -> ResourceFolder:
        return self.context.resource_folder


    @property
    def settings(self) -> TSettings:
        return self.context.settings

    # endregion

    # region Convenience methods
    def is_installed(self):
        from .controller_registry import ControllerRegistry
        registry = ControllerRegistry([self])
        status = registry.get_installation_statuses()[self.get_name()]
        return status.installed


    def get_name(self):
        from .controller_registry import ControllerRegistry
        return ControllerRegistry.to_controller_name(type(self))


    def get_snake_case_name(self):
        name = self.get_name()
        result = []
        for i, c in enumerate(name):
            if c.islower():
                result.append(c)
                continue
            if i > 0 and name[i-1].islower():
                result.append('_')
            result.append(c.lower())
        return ''.join(result)


    def is_running(self, parameter: str|None):
        return parameter in self.get_running_instances_id_to_parameter().values()

    def is_any_running(self):
        return len(self.get_running_instances_id_to_parameter()) != 0

    def reinstall(self, purge: bool = False):
        if self.is_installed():
            self.uninstall(purge)
        self.install()

    def stop_all(self):
        for instance_id in self.get_running_instances_id_to_parameter():
            self.stop(instance_id)

    def _run_self_test(self, api, tc: TestCase, locator: Locator):
        output_folder = locator.self_test_path
        html_path = output_folder / (self.get_name() + '.html')
        if str(html_path).startswith('/'):
            html_link = f'file://{html_path}'
        else:
            html_link = f'file:///{html_path}'.replace('\\','/')
        items = []
        error = None
        try:
            for item in self._self_test_internal(api, tc):
                items.append(item)
        except Exception as ex:
            error = traceback.format_exc()
            raise ValueError(f"Exception has occured during self-test. View the report {html_link}") from ex
        finally:
            report = TestReport(self.get_name(), items, error, type(self))
            FileIO.write_pickle(report, output_folder/self.get_name())
            FileIO.write_text(create_self_test_report_page(report), html_path)
        if error is None:
            print(f"SELF-TEST EXITED SUCCESSFULLY. Report {html_link}")
        return report


    def self_test(self,
                  tc: None|TestCase = None,
                  locator: Locator = Loc,
                  api = None
                  ) -> TestReport:
        if tc is None:
            tc = TestCase()

        from brainbox.framework import BrainBoxApi
        if api is not None:
            if not isinstance(api, BrainBoxApi):
                raise ValueError("Api should be BrainBoxApi")
            else:
                return self._run_self_test(api, tc, locator)

        with BrainBoxApi.Test(port=18091, locator=locator, run_controllers_in_default_environment=False) as api:
            try:
                result = self._run_self_test(api, tc, locator)
                return result
            finally:
                self.stop_all()







    # endregion


from pathlib import Path
import sys
from dataclasses import dataclass

train_path = Path('/models')
rhasspy_path = Path('/profiles')
profile_path = Path('/usr/lib/rhasspy/rhasspy-profile/rhasspyprofile/profiles/')
kaldi_path = Path('/usr/lib/rhasspy/.venv/lib/kaldi/')


def load_rhasspy_modules():
    dir_file = Path('/usr/lib/rhasspy/RHASSPY_DIRS')
    dirs = []
    with open(dir_file, 'r') as file:
        while True:
            d = file.readline().strip()
            if d == '':
                break
            dirs.append(d)

    for d in dirs:
        sys.path.append(str(dir_file.parent / d))

